program xsltio;

{$APPTYPE CONSOLE}

uses

  Classes, SysUtils, libxml2, libxslt;

var
  resourceStream: TFileStream;

(**
 * customIO_evaluate:
 * @URI: an URI to test
 *
 * Returns 1 if yes and 0 if another Input module should be used
 *)
function customIO_evaluate(aURI: pchar): longint; cdecl;
// Type: xmlInputMatchCallback
// Evaluates if the resource should be loaded using the custom loading machanism
var
  p, pEnd: pxmlChar;

begin
  if (aURI <> nil) and (xmlStrncmp(aURI, 'urn:', 4) = 0) then begin
    // The "URN" scheme of the URI was found.
    { A result of 1 indicates that Libxml2 should use the registered CUSTOM
      loader. }
    result := 1;
    // EXIT.
    exit;
  end;
  { A result of zero indicates that Libxml2 should use the DEFAULT loading
    machanism. }
  result := 0;
end;

(**
 * customIO_openResource:
 * @URI: an URI to test
 *
 * Returns an Input context or NULL in case or error
 *)
function customIO_openResource(aURI: pchar): pointer; cdecl;
// Type: xmlInputOpenCallback
begin
  if (aURI = nil) or (xmlStrncmp(aURI, 'urn:', 4) <> 0) then begin
    result := nil;
    // EXIT.
    exit;
  end;
  // Set the context; it will be the resource stream in this case.
  result := pointer(resourceStream);
end;

(**
 * customIO_closeResource:
 * @context: the read context
 *
 * Close
 *
 * Returns 0 or -1 in case of error
 *)
function customIO_closeResource(aContext: pointer): longint; cdecl;
// Type: xmlInputCloseCallback
// Finalizes the resource loading.
begin
  if (aContext = nil) then begin
    // Error.
    result := -1;
    // EXIT.
    exit;
  end;
  result := 0;
  { TODO: Free the resource stream here? Maby not robust enough, since the transformation
    process could be killed before this one is called. }
end;

(**
 * customIO_readResource:
 * @context: the read context
 * @buffer: where to store data
 * @len: number of bytes to read
 *
 * read.
 *
 * Returns the number of bytes read or -1 in case of error
 *)
function customIO_readResource(aContext: pointer; aBuffer: pchar; aLength: integer): longint; cdecl;
// Type: xmlInputReadCallback
begin
  // Check arguments.
  if ((aContext = nil) or (aBuffer = nil) or (aLength < 0)) then begin
    // Error: bad arguments.
    // TODO: raise an exception?
    result := -1;
    // EXIT.
    exit;
  end;
  { Note that in this example the 'context' will be a pointer to the resourceStream. }
  // Write the chunk into the buffer.
  with TFileStream(aContext) do begin
    if (Size - Position = 0) then begin
      result := 0;
      // EXIT.
      exit;
    end else if (Size - Position < aLength) then
      aLength := Size - Position;
    result := TFileStream(aContext).Read(aBuffer^, aLength);
  end;
end;

function transform(aDoc: xmlDocPtr; var aStylesheetDoc: xmlDocPtr): xmlDocPtr;
var
  stylesheetDoc: xmlDocPtr;
  parsedStylesheet: xsltStylesheetPtr;
  transformCtxt: xsltTransformContextPtr;
begin
  result := nil;
  // Parse the stylesheet document.
  parsedStylesheet := xsltParseStyleSheetDoc(aStylesheetDoc);
  if (parsedStylesheet = nil) then
    raise Exception.create('Parsing of stylesheet document failed.');
  try
    // Create the transformation context.
    transformCtxt := xsltNewTransformContext(parsedStylesheet, aDoc);
    try
      // Transform.
      result := xsltApplyStylesheetUser(parsedStylesheet, aDoc, nil, nil, nil, transformCtxt);
      if (result = nil) then
        raise Exception.create('Transformation of the document failed.');
    finally
      xsltFreeTransformContext(transformCtxt);
    end;
  finally
    { Consume the stylesheet document, since it will be freed when the
      parsed stylesheet is freed. } 
    aStylesheetDoc := nil;
    xsltFreeStylesheet(parsedStylesheet);
  end;
end;

function loadDoc(aURI: pchar): xmlDocPtr;
var
  res: integer;
  parserCtxt: xmlParserCtxtPtr;
begin
  result := nil;
  parserCtxt := xmlCreateFileParserCtxt(aURI); // xmlCreateDocParserCtxt(aXmlText);
  try
    // Turn of DTD validation (default).
    parserCtxt.validate := 0;
    // Turn of entity substitution.
    parserCtxt.replaceEntities := 0;
    // Don't remove whitespace content of elements of 'mixed' type.
    parserCtxt.keepBlanks := 1;
    { If no XML declaration was used, libxml2 tries to detect the encoding
      using the first 4 bytes of the text; if this does not work it assumes the
      text to be UTF-8 encoded. If the parser encounters an UTF-8 error it
      falls back to Latin-1 encoding.
    }
    // Parse.
    res := xmlParseDocument(parserCtxt);
    if (res <> 0) or (parserCtxt.wellFormed = 0) then
    begin
      // Failed. Free the doc if still existent.
      if (parserCtxt.myDoc <> nil) then begin
        xmlFreeDoc(parserCtxt.myDoc);
        parserCtxt.myDoc := nil;
      end;
      // Exception
      raise Exception.Create('Failed to parse the document.');
    end else
      // Succeeded.
      result := parserCtxt.myDoc;
  finally
    xmlFreeParserCtxt(parserCtxt);
  end;
end;

procedure test;
var
  doc, stylesheetDoc, resultDoc: xmlDocPtr;
begin
  // Register the custom I/O handlers.
  if (xmlRegisterInputCallbacks(customIO_evaluate, customIO_openResource, customIO_readResource, customIO_closeResource) < 0) then
    raise Exception.Create('Failed to register the custom I/O handlers.');
  doc := loadDoc(pchar('..\data\document.xml'));
  stylesheetDoc := loadDoc(pchar('..\data\stylesheet.xsl'));
  try
    resultDoc := transform(doc, stylesheetDoc);
    // Save the doc.
    xmlSaveFormatFileEnc(pchar('..\data\result.xml'), resultDoc, 'ISO-8859-1', 1);

  finally
    xmlFreeDoc(doc);
    if (stylesheetDoc <> nil) then
      xmlFreeDoc(stylesheetDoc);
  end;

end;

(********
* MAIN. *
********)

begin
  xmlInitParser();
  resourceStream := TFileStream.Create('..\data\resource.xml', fmOpenRead);
  try
    test();
  finally
    resourceStream.Free;
  end;
  xmlCleanupParser();
end.
